
#ifdef HAVE_CONFIG_H
#include "config.h"
#endif /* HAVE_CONFIG_H */
#include "interface.h"
#include <stdio.h>
#include <string.h>

#define AU_LAME

#ifdef AU_LAME

#ifdef HAVE_UNISTD_H
#include <unistd.h>
#endif /* HAVE_UNISTD_H */

#ifdef STDC_HEADERS
#include <string.h>
#include <stdlib.h>
#include <ctype.h>
#include <time.h>
#elif HAVE_STRINGS_H
#include <strings.h>
#endif
#ifdef HAVE_FCNTL_H
#include <fcntl.h>
#endif

#ifdef __W32__
#include <windows.h>
#include <winnls.h>
#endif

#include <vorbis/vorbisenc.h>

#include "timidity.h"
#include "common.h"
#include "output.h"
#include "controls.h"
#include "instrum.h"
#include "playmidi.h"
#include "readmidi.h"

static int open_output(void); /* 0=success, 1=warning, -1=fatal error */
static void close_output(void);
static int output_data(char *buf, int32 bytes);
static int acntl(int request, void *arg);

#include "LameEnc.h"
#include "BladeMP3EncDLL.h"

/* export the playback mode */
#define dpm lame_play_mode

PlayMode dpm = {
    44100, PE_16BIT|PE_SIGNED, PF_PCM_STREAM|PF_FILE_OUTPUT,
    -1,
    {0,0,0,0,0},
    "Lame", 'L',
    NULL,
    open_output,
    close_output,
    output_data,
    acntl
};
static char *tag_title = NULL;

#if defined ( IA_W32GUI ) || defined ( IA_W32G_SYN )
extern char *w32g_output_dir;
extern int w32g_auto_output_mode;
#endif

/*************************************************************************/

#if defined ( IA_W32GUI ) || defined ( IA_W32G_SYN )
/*
static int
choose_bitrate(int nch, int rate)
{
  int bitrate;

	if (ogg_vorbis_mode < 1 || ogg_vorbis_mode > 1000)
		bitrate = 8;
	else
		bitrate = ogg_vorbis_mode;
	return bitrate;

  return (int)(nch * rate * (128000.0 / (2.0 * 44100.0)) + 0.5);
}
*/
#else
/*
static int
choose_bitrate(int nch, int rate)
{
  int target;

  target = (int)(nch * rate * (128000.0 / (2.0 * 44100.0)) + 0.5);

  return target;
}
*/
#endif

#ifdef __W32__
static char *w32_mbs_to_utf8(const char* str)
{
	int str_size = strlen(str);
	int buff16_size = str_size;
	wchar_t* buff16;
	int buff8_size = 0;
	char* buff8;
	char* buff8_p;
	int i;
	if ( str_size == 0 ) {
		return strdup ( str );
	}
	buff16 = (wchar_t*) malloc (sizeof(wchar_t)*buff16_size + 1);
	if ( buff16 == NULL ) return NULL;
	buff16_size = MultiByteToWideChar(CP_OEMCP, MB_PRECOMPOSED, str, str_size, buff16, buff16_size ) ;
	if ( buff16_size == 0 ) {
		free ( buff16 );
		return NULL;
	}
	for ( i = 0; i < buff16_size; ++i ) {
		wchar_t w = buff16[i];
		if ( w < 0x0080 ) buff8_size += 1;
		else if ( w < 0x0800 ) buff8_size += 2;
		else buff8_size += 3;
	}
	buff8 = (char*) malloc ( sizeof(char)*buff8_size + 1 );
	if ( buff8 == NULL ) {
		free ( buff16 );
		return NULL;
	}
	for ( i = 0, buff8_p = buff8; i < buff16_size; ++i ) {
		wchar_t w = buff16[i];
		if ( w < 0x0080 ) {
			*(buff8_p++) = (char)w;
		} else if ( buff16[i] < 0x0800 ) {
			*(buff8_p++) = 0xc0 | (w >> 6);
			*(buff8_p++) = 0x80 | (w & 0x3f);
		} else {
			*(buff8_p++) = 0xe0 | (w >> 12);
			*(buff8_p++) = 0x80 | ((w >>6 ) & 0x3f);
			*(buff8_p++) = 0x80 | (w & 0x3f);
		}
	}
	*buff8_p = '\0';
	free ( buff16 );
	return buff8;
}
#endif

HMODULE hLame;
BEINITSTREAM		beInitStream=NULL;
BEENCODECHUNK		beEncodeChunk=NULL;
BEDEINITSTREAM		beDeinitStream=NULL;
BECLOSESTREAM		beCloseStream=NULL;
BEVERSION			beVersion=NULL;
BEWRITEVBRHEADER	beWriteVBRHeader=NULL;
BEWRITEINFOTAG		beWriteInfoTag=NULL;

static DWORD dwSamples = 0;
static HBE_STREAM	hbeStream = 0;
static FILE *lame_fpout = NULL;
unsigned char *lame_buffer = NULL;
unsigned char *lame_work_buffer = NULL;
static int lame_buffer_cur = 0;

typedef enum  {
	INSANE, 
	FAST_EXTREME, 
	EXTREME, 
	FAST_STANDARD, 
	STANDARD, 
	FAST_MEDIUM, 
	MEDIUM,
	CBR320, 
	CBR256,
	CBR224,
	CBR192, 
	CBR160, 
	CBR128, 
	CBR112, 
	CBR96, 
	CBR80,
	CBR64,

	QUALITY_SETTING_LAST = MEDIUM, 
	CBR_SETTING_LAST = CBR64, 
}LAME_PRESET_TIM;

int lame_encode_preset = 0;

static int lame_output_open(const char *fname, const char *comment)
{
  int fd = 1;
  int nch;
	FILE*		pFileIn			=NULL;
	FILE*		pFileOut		=NULL;
	BE_VERSION	Version			={0,};
	BE_CONFIG	beConfig		={0,};

	CHAR		strFileIn[255]	={'0',};
	CHAR		strFileOut[255]	={'0',};

	DWORD		dwMP3Buffer		=0;
	BE_ERR		err				=0;

	PBYTE		pMP3Buffer		=NULL;
	PSHORT		pWAVBuffer		=NULL;

#if !defined ( IA_W32GUI ) && !defined ( IA_W32G_SYN )
  int bitrate;
#endif

	hLame = LoadLibrary("lame_enc.dll");
  	if ( NULL == hLame ){
		return -1;
	}


	beInitStream	= (BEINITSTREAM) GetProcAddress(hLame, TEXT_BEINITSTREAM);
	beEncodeChunk	= (BEENCODECHUNK) GetProcAddress(hLame, TEXT_BEENCODECHUNK);
	beDeinitStream	= (BEDEINITSTREAM) GetProcAddress(hLame, TEXT_BEDEINITSTREAM);
	beCloseStream	= (BECLOSESTREAM) GetProcAddress(hLame, TEXT_BECLOSESTREAM);
	beVersion		= (BEVERSION) GetProcAddress(hLame, TEXT_BEVERSION);
	beWriteVBRHeader= (BEWRITEVBRHEADER) GetProcAddress(hLame,TEXT_BEWRITEVBRHEADER);
	beWriteInfoTag  = (BEWRITEINFOTAG) GetProcAddress(hLame,TEXT_BEWRITEINFOTAG);

	if(!beInitStream || !beEncodeChunk || !beDeinitStream || !beCloseStream || !beVersion || !beWriteVBRHeader){
		return -1;
	}

	nch = (dpm.encoding & PE_MONO) ? 1 : 2;

	beVersion( &Version );

#if defined ( IA_W32GUI ) || defined ( IA_W32G_SYN )
//  lame_ConfigDialogInfoApply();
#endif
	memset(&beConfig,0,sizeof(beConfig));					// clear all fields

	// use the LAME config structure
	beConfig.dwConfig = BE_CONFIG_LAME;

	// this are the default settings for testcase.wav
	beConfig.format.LHV1.dwStructVersion	= 1;
	beConfig.format.LHV1.dwStructSize		= sizeof(beConfig);		
	beConfig.format.LHV1.dwSampleRate		= play_mode->rate;				// INPUT FREQUENCY
	beConfig.format.LHV1.dwReSampleRate		= 0;					// DON"T RESAMPLE
	beConfig.format.LHV1.nMode				= (nch == 1) ? BE_MP3_MODE_MONO : BE_MP3_MODE_JSTEREO;	// OUTPUT IN STREO
	
	if (lame_encode_preset <= QUALITY_SETTING_LAST) { 
		switch (lame_encode_preset) {
		case INSANE:
			beConfig.format.LHV1.nPreset = LQP_INSANE;
			break;
		case FAST_EXTREME:
			beConfig.format.LHV1.nPreset = LQP_FAST_EXTREME;
			break;
		case EXTREME:
			beConfig.format.LHV1.nPreset = LQP_EXTREME;
			break;
		case FAST_STANDARD:
			beConfig.format.LHV1.nPreset = LQP_FAST_STANDARD;
			break;
		case STANDARD:
			beConfig.format.LHV1.nPreset = LQP_STANDARD;
			break;
		case FAST_MEDIUM:
			beConfig.format.LHV1.nPreset = LQP_FAST_MEDIUM;
			break;
		case MEDIUM:
			beConfig.format.LHV1.nPreset = LQP_MEDIUM;
			break;
		}
		beConfig.format.LHV1.bWriteVBRHeader	= TRUE;
		beConfig.format.LHV1.nVbrMethod			= VBR_METHOD_NEW;
		beConfig.format.LHV1.bEnableVBR			= TRUE;
	}else{ // cbr
		switch (lame_encode_preset) {
		case CBR320:
			beConfig.format.LHV1.dwBitrate = beConfig.format.LHV1.dwMaxBitrate = 320;
			break;
		case CBR256:
			beConfig.format.LHV1.dwBitrate = beConfig.format.LHV1.dwMaxBitrate = 256;
			break;
		case CBR224:
			beConfig.format.LHV1.dwBitrate = beConfig.format.LHV1.dwMaxBitrate = 224;
			break;
		case CBR192:
			beConfig.format.LHV1.dwBitrate = beConfig.format.LHV1.dwMaxBitrate = 192;
			break;
		case CBR160:
			beConfig.format.LHV1.dwBitrate = beConfig.format.LHV1.dwMaxBitrate = 160;
			break;
		case CBR128:
			beConfig.format.LHV1.dwBitrate = beConfig.format.LHV1.dwMaxBitrate = 128;
			break;
		case CBR112:
			beConfig.format.LHV1.dwBitrate = beConfig.format.LHV1.dwMaxBitrate = 112;
			break;
		case CBR96:
			beConfig.format.LHV1.dwBitrate = beConfig.format.LHV1.dwMaxBitrate = 96;
			break;
		case CBR80:
			beConfig.format.LHV1.dwBitrate = beConfig.format.LHV1.dwMaxBitrate = 80;
			break;
		case CBR64:
			beConfig.format.LHV1.dwBitrate = beConfig.format.LHV1.dwMaxBitrate = 64;
			break;
		default:
			beConfig.format.LHV1.dwBitrate = beConfig.format.LHV1.dwMaxBitrate = 128;
			break;
		}
	}

	beConfig.format.LHV1.dwMpegVersion		= MPEG1;				// MPEG VERSION (I or II)
	beConfig.format.LHV1.dwPsyModel			= 0;					// USE DEFAULT PSYCHOACOUSTIC MODEL 
	beConfig.format.LHV1.dwEmphasis			= 0;					// NO EMPHASIS TURNED ON
	beConfig.format.LHV1.bOriginal			= TRUE;					// SET ORIGINAL FLAG
	beConfig.format.LHV1.bWriteVBRHeader	= TRUE;					// Write INFO tag

	beConfig.format.LHV1.bCRC				= TRUE;					// INSERT CRC
//	beConfig.format.LHV1.bCopyright			= TRUE;					// SET COPYRIGHT FLAG	
	beConfig.format.LHV1.bPrivate			= TRUE;					// SET PRIVATE FLAG
	beConfig.format.LHV1.bNoRes				= TRUE;					// No Bit resorvoir

	// Init the MP3 Stream
	err = beInitStream(&beConfig, &dwSamples, &dwMP3Buffer, &hbeStream);

	// Check result
	if(err != BE_ERR_SUCCESSFUL){
		fprintf(stderr,"Error opening encoding stream (%lu)", err);
		return -1;
	}

#if !defined ( IA_W32GUI ) && !defined ( IA_W32G_SYN )

#else

#endif


  /* add default tag */
    if (tag_title != NULL) {
#ifndef __W32__

#else

#endif
  }

	lame_fpout = fopen(fname, "wb");
	lame_buffer = (unsigned char  *)malloc(dwSamples * 8);
	lame_work_buffer = (unsigned char  *)malloc(dwSamples * 8);
	lame_buffer_cur = 0;

  return fd;
}

/* mode
  0,1: Default mode.
  2: Remove the directory path of input_filename, then add output_dir.
  3: Replace directory separator characters ('/','\',':') with '_', then add output_dir.
 */
extern char *create_auto_output_name(const char *input_filename, char *ext_str, char *output_dir, int mode);

static int auto_lame_output_open(const char *input_filename, const char *title)
{
  char *output_filename;

#if !defined ( IA_W32GUI ) && !defined ( IA_W32G_SYN )
  output_filename = create_auto_output_name(input_filename,"ogg",NULL,0);
#else
  output_filename = create_auto_output_name(input_filename,"mp3",w32g_output_dir,w32g_auto_output_mode);
#endif
  if(output_filename==NULL){
	  return -1;
  }
  if (tag_title != NULL) {
	free(tag_title);
	tag_title = NULL;
  }
  if (title != NULL) {
	tag_title = (char *)safe_malloc(sizeof(char)*(strlen(title)+1));
	strcpy(tag_title, title);
  }
  if((dpm.fd = lame_output_open(output_filename, input_filename)) == -1) {
    free(output_filename);
    return -1;
  }
  if(dpm.name != NULL)
    free(dpm.name);
  dpm.name = output_filename;
  ctl->cmsg(CMSG_INFO, VERB_NORMAL, "Output %s", dpm.name);
  return 0;
}

static int open_output(void)
{
  int include_enc, exclude_enc;

  /********** Encode setup ************/

  include_enc = exclude_enc = 0;

  /* only 16 bit is supported */
  include_enc |= PE_16BIT|PE_SIGNED;
  exclude_enc |= PE_BYTESWAP|PE_24BIT;
  dpm.encoding = validate_encoding(dpm.encoding, include_enc, exclude_enc);

#if !defined ( IA_W32GUI ) && !defined ( IA_W32G_SYN )
  if(dpm.name == NULL) {
    dpm.flag |= PF_AUTO_SPLIT_FILE;
  } else {
    dpm.flag &= ~PF_AUTO_SPLIT_FILE;
    if((dpm.fd = lame_output_open(dpm.name, NULL)) == -1)
      return -1;
  }
#else
	if(w32g_auto_output_mode>0){
      dpm.flag |= PF_AUTO_SPLIT_FILE;
      dpm.name = NULL;
    } else {
      dpm.flag &= ~PF_AUTO_SPLIT_FILE;
      if((dpm.fd = lame_output_open(dpm.name,NULL)) == -1)
		return -1;
    }
#endif

  return 0;
}

static int output_data(char *readbuffer, int32 bytes)
{
	BE_ERR err;
	DWORD dwRead = 0, dwWrite;
	int read_size = 0;
	int sample_sz = dwSamples*2;
	int total_pos = 0;
	int bytes_tmp = bytes; // ǂݍނтɃfN
/*
	while ( (bytes_tmp + lame_buffer_cur) > sample_sz) {
		read_size = abs(bytes_tmp - lame_buffer_cur);

		if (read_size > sample_sz) {
			read_size = abs((int)dwSamples - lame_buffer_cur);
			total_pos += read_size;
			bytes_tmp -= read_size;
		}else{
			total_pos += read_size;
			bytes_tmp -= read_size;
		}
		dwRead = read_size;

		memcpy(lame_buffer + lame_buffer_cur, readbuffer + total_pos, read_size);


		err = beEncodeChunk(hbeStream, dwRead/2, (short *)lame_buffer, (short *)lame_work_buffer, &dwWrite);
		if(err != BE_ERR_SUCCESSFUL){
			beCloseStream(hbeStream);
			return -1;
		}
		if(fwrite(lame_work_buffer, 1, dwWrite, lame_fpout) != dwWrite){
			return -1;
		}

		lame_buffer_cur = 0;
	}

	read_size = abs( (bytes - total_pos) - lame_buffer_cur);
	memcpy(lame_buffer + lame_buffer_cur, readbuffer, read_size);
	lame_buffer_cur += read_size;
*/

	err = beEncodeChunk(hbeStream, bytes>>1, (short *)readbuffer, (short *)lame_work_buffer, &dwWrite);
	if(err != BE_ERR_SUCCESSFUL){
		beCloseStream(hbeStream);
		return -1;
	}
	if(fwrite(lame_work_buffer, 1, dwWrite, lame_fpout) != dwWrite){
		return -1;
	}

	return 0;
}

static void close_output(void)
{
	BE_ERR err;
	DWORD dwRead, dwWrite;

	if (hbeStream) {

		err = beDeinitStream(hbeStream, lame_work_buffer, &dwWrite);

		// Check result
		if(err != BE_ERR_SUCCESSFUL){
			return;
		}

		if( dwWrite ){
			if( fwrite( lame_work_buffer, 1, dwWrite, lame_fpout ) != dwWrite ){
			}
		}

		beCloseStream( hbeStream );

		free(lame_buffer);
		free(lame_work_buffer);
		FreeLibrary(hLame);
		fclose(lame_fpout);

	}

	hbeStream = 0;


	dpm.fd = -1;
}

static int acntl(int request, void *arg)
{
  switch(request) {
  case PM_REQ_PLAY_START:
    if(dpm.flag & PF_AUTO_SPLIT_FILE){
      if(  ( NULL == current_file_info ) || (NULL == current_file_info->filename ) )
        return auto_lame_output_open("Output.mid",NULL);
      return auto_lame_output_open(current_file_info->filename, current_file_info->seq_name);
   }
    return 0;
  case PM_REQ_PLAY_END:
    if(dpm.flag & PF_AUTO_SPLIT_FILE)
      close_output();
    return 0;
  case PM_REQ_DISCARD:
    return 0;
  }
  return -1;
}

#include "resource.h"
#include <windowsx.h>

extern char *timidity_window_inifile;
extern char *timidity_output_inifile;

#if defined IA_W32GUI || defined( IA_W32G_SYN )

#define SEC_LAME	"lame"
void lame_ConfigDialogInfoSaveINI(void)
{
	char *section = SEC_LAME;
	char *inifile = timidity_output_inifile;
	char buffer[1024];
#if defined(__MINGW32__) || defined(__CYGWIN__)
#define NUMSAVE(name) \
		sprintf(buffer,"%d",gogo_ConfigDialogInfo.name ); \
		WritePrivateProfileString(section, #name ,buffer,inifile);
#define STRSAVE(name) \
		WritePrivateProfileString(section,(char *) #name ,(char *)gogo_ConfigDialogInfo.name ,inifile);
#else
#define NUMSAVE(name) \
		sprintf(buffer,"%d",name ); \
		WritePrivateProfileString(section, #name ,buffer,inifile);
#define STRSAVE(name) \
		WritePrivateProfileString(section,(char *) #name ,(char *)name ,inifile);
#endif
	NUMSAVE(lame_encode_preset)
	WritePrivateProfileString(NULL,NULL,NULL,inifile);		// Write Flush
#undef NUMSAVE
#undef STRSAVE
	return;
}

void lame_ConfigDialogInfoLoadINI(void)
{
	char *inifile = timidity_output_inifile;
	char *section = SEC_LAME;
	int num;
#define NUMLOAD(name) \
		num = GetPrivateProfileInt(section, #name ,-1,inifile); \
		if(num!=-1) name = num;

	NUMLOAD(lame_encode_preset)
#undef NUMLOAD
}

#endif

LRESULT WINAPI lameConfigDialogProc(HWND hwnd, UINT msg, WPARAM wp, LPARAM lp)
{
	switch (msg) {
		case WM_INITDIALOG:
		{
			HWND hcb = GetDlgItem(hwnd, IDC_LAME_CBPRESET);
			ComboBox_AddString(hcb, "preset: insane");
			ComboBox_AddString(hcb, "preset: fast extreme");
			ComboBox_AddString(hcb, "preset: extreme");
			ComboBox_AddString(hcb, "preset: fast standard");
			ComboBox_AddString(hcb, "preset: standard");
			ComboBox_AddString(hcb, "preset: fast medium");
			ComboBox_AddString(hcb, "preset: medium");

			ComboBox_AddString(hcb, "CBR 320 kbps");
			ComboBox_AddString(hcb, "CBR 256 kbps");
			ComboBox_AddString(hcb, "CBR 224 kbps");
			ComboBox_AddString(hcb, "CBR 192 kbps");
			ComboBox_AddString(hcb, "CBR 160 kbps");
			ComboBox_AddString(hcb, "CBR 128 kbps");
			ComboBox_AddString(hcb, "CBR 112 kbps");
			ComboBox_AddString(hcb, "CBR 96 kbps");
			ComboBox_AddString(hcb, "CBR 80 kbps");
			ComboBox_AddString(hcb, "CBR 64 kbps");

			ComboBox_SetCurSel(hcb, lame_encode_preset);
			return TRUE;
		}
		case WM_CLOSE:
			EndDialog(hwnd,TRUE);
			break;
		case WM_COMMAND:
			switch (LOWORD(wp)) {
				case IDCANCEL:
					EndDialog(hwnd,TRUE);
					break;
				case IDC_LAME_CBPRESET:
					break;
				case IDOK:
					lame_encode_preset = ComboBox_GetCurSel( GetDlgItem(hwnd, IDC_LAME_CBPRESET) );
#ifdef IA_W32GUI
					lame_ConfigDialogInfoSaveINI();
#endif
					EndDialog(hwnd,TRUE);
					break;
			}
			break;
	}
	return FALSE;
}

void lameConfigDialog(HWND hwnd)
{
	DialogBox(GetModuleHandle(NULL), MAKEINTRESOURCE(IDD_DIALOG_LAME), hwnd, (DLGPROC)lameConfigDialogProc);
}


#endif
