#ifdef HAVE_CONFIG_H
#include "config.h"
#endif /* HAVE_CONFIG_H */

#ifndef NO_STRING_H
#include <string.h>
#else
#include <strings.h>
#endif
#include "timidity.h"
#include "controls.h"
#include "tables.h"
#include "common.h"
#include "output.h"
#define REVERB_PRIVATE 1
#include "reverb.h"
#include "mt19937ar.h"
#include <math.h>
#include <stdlib.h>
#include "sysdep.h"

extern PlayMode *play_mode;

extern OVERRIDETIMIDITYDATA otd;

HMODULE hVSTHost = NULL;

extern void compressor
    (
        int32*  wav_in,     // signal
        int     n,          // N samples
        double  threshold,  // threshold (percents)
        double  slope,      // slope angle (percents)
        double  sr,         // sample rate (smp/sec)
        double  tla,        // lookahead  (ms)
        double  twnd,       // window time (ms)
        double  tatt,       // attack time  (ms)
        double  trel        // release time (ms)
    );


static inline int clip_int(int val, int min, int max)
{
	return ((val > max) ? max : (val < min) ? min : val);
}

void do_shelving_filter_stereo(int32* buf, int32 count, filter_shelving *p)
{
	int32 i;
	int32 x1l = p->x1l, x2l = p->x2l, y1l = p->y1l, y2l = p->y2l,
		x1r = p->x1r, x2r = p->x2r, y1r = p->y1r, y2r = p->y2r, yout;
	int32 a1 = p->a1, a2 = p->a2, b0 = p->b0, b1 = p->b1, b2 = p->b2;

	for(i = 0; i < count; i++) {
		yout = imuldiv24(buf[i], b0) + imuldiv24(x1l, b1) + imuldiv24(x2l, b2) + imuldiv24(y1l, a1) + imuldiv24(y2l, a2);
		x2l = x1l;
		x1l = buf[i];
		y2l = y1l;
		y1l = yout;
		buf[i] = yout;

		yout = imuldiv24(buf[++i], b0) + imuldiv24(x1r, b1) + imuldiv24(x2r, b2) + imuldiv24(y1r, a1) + imuldiv24(y2r, a2);
		x2r = x1r;
		x1r = buf[i];
		y2r = y1r;
		y1r = yout;
		buf[i] = yout;
	}
	p->x1l = x1l, p->x2l = x2l, p->y1l = y1l, p->y2l = y2l,
		p->x1r = x1r, p->x2r = x2r, p->y1r = y1r, p->y2r = y2r;
}

void conv_gs_stereo_enhancer(struct insertion_effect_gs_t *ieffect, EffectList *ef)
{
	InfoStereoEQ *eq = (InfoStereoEQ *)ef->info;
	InfoEnhancer *enhan = (InfoEnhancer *)ef->info;
/*
	eq->high_freq = 8000;
	eq->high_gain = 12 + clip_int(ieffect->parameter[12] - 0x34, -12, 12);
	eq->low_freq = 800;
	eq->low_gain =  -12 + clip_int(ieffect->parameter[11] - 0x34, -12, 12);
*/
	eq->low_freq = 200;
	eq->low_gain = -6 + clip_int(ieffect->parameter[11] - 0x34, -12, 12);
	eq->high_freq = 4000;
	eq->high_gain = 6 + clip_int(ieffect->parameter[12] - 0x34, -12, 12);
	eq->m1_freq = 6000;
	eq->m1_q = 4.0;
	eq->m1_gain = 8;
	eq->m2_freq = 6000;
	eq->m2_q = 4.0;
	eq->m2_gain = 8;
	eq->level = (double)127.0 / 127.0;

	enhan->level_enhance = 1.0;
}

void do_stereo_enhancer(int32 *buf, int32 count, EffectList *ef)
{
	double mono;
	double tmpL, tmpR;
	double delta;
	int32 i;
	double outL, outR;
	double coef = 0.4;
	InfoEnhancer *info = (InfoEnhancer *)ef->info;
	InfoStereoEQ *eq = (InfoStereoEQ *)ef->info;

	

	if(count == MAGIC_INIT_EFFECT_INFO) {
		do_stereo_eq(buf, count, ef);
		return;
	} else if(count == MAGIC_FREE_EFFECT_INFO) {
		do_stereo_eq(buf, count, ef);
		return;
	}

	do_stereo_eq(buf, count, ef);

	for (i = 0; i < count; ++i) {

		tmpL = buf[i];
		tmpR = buf[i+1];

		mono = (tmpL + tmpR)/2.0;
		delta = tmpL - mono;
		delta = delta * coef;

		outL = tmpL + delta;
		outR = tmpR - delta;

		buf[i] = outL * info->level_enhance;
		buf[i+1] = outR * info->level_enhance;

		++i;
	}
}

void mix_compressor(int32 *buf, int count)
{
	if (otd.timRunMode & EOWM_USE_COMPRESSOR)
		compressor(buf, count, 
		otd.compThr, otd.compSlope, play_mode->rate, otd.compLook, 
		otd.compWTime, otd.compATime, otd.compRTime);
}


int32 Sample_FtoI(double v)
{
	if (v<0)return ( (-v * (double)LONG_MIN) - 0.5 );
	return (v * (double)LONG_MAX + 0.5);
}

double Sample_ItoF(int32 v)
{
	if (v<0)return (-(double)v / (double)LONG_MIN);
	return ((double)v / (double)LONG_MAX);
}


void compressor
(
	int32*  wav_in,     // signal
	int     n,          // N samples
	double  threshold,  // threshold (percents)
	double  slope,      // slope angle (percents)
	double  sr,         // sample rate (smp/sec)
	double  tla,        // lookahead  (ms)
	double  twnd,       // window time (ms)
	double  tatt,       // attack time  (ms)
	double  trel        // release time (ms)
)
{
	double att, rel, env, rms, theta=0, gain=0, summ;
	int lhsmp=0, nrms=0, i, j;
	typedef int32 wavlr[2];
	wavlr *wav = (wavlr *)wav_in;
	
	n=n/2;

	threshold *= 0.01;          // threshold to unity (0...1)
	slope *= 0.01;              // slope to unity
	tla   *= 0.001;                // lookahead time to seconds
	twnd  *= 0.001;               // window time to seconds
	tatt  *= 0.001;               // attack time to seconds
	trel  *= 0.001;               // release time to seconds

	// attack and release "per sample decay"
	att = (tatt == 0.0) ? (0.0) : exp (-1.0 / (sr * tatt));
	rel = (trel == 0.0) ? (0.0) : exp (-1.0 / (sr * trel));

	// envelope
	env = 0.0;

	// sample offset to lookahead wnd start
	lhsmp = (int)(sr * tla);

	// samples count in lookahead window
	nrms = (int)(sr * twnd);

	// for each sample...
	for (i = 0; i < n; ++i)
	{
		// now compute RMS
		summ = 0;

		// for each sample in window
		for (j = 0; j < nrms; ++j)
		{
			int lki = i + j + lhsmp;
			if (lki < n) {
				double  smp = Sample_ItoF((wav[lki][0] + wav[lki][1])/2);
				summ += smp * smp;  // square em..
			}else{
				break;
			}
		}


		rms = sqrt(summ / (double)nrms);   // root-mean-square

		// dynamic selection: attack or release?
		theta = rms > env ? att : rel;

		// smoothing with capacitor, envelope extraction...
		// here be aware of pIV denormal numbers glitch
		env = (1.0 - theta) * rms + theta * env;

		// the very easy hard knee 1:N compressor
		gain = 1.0;
		if (env > threshold)
			gain = gain - (env - threshold) * slope;

		// result - two hard kneed compressed channels... 
		wav[i][0] = Sample_FtoI( Sample_ItoF(wav[i][0]) * gain);
		wav[i][1] = Sample_FtoI( Sample_ItoF(wav[i][1]) * gain);
	}
}

