/*
    TiMidity++ -- MIDI to WAVE converter and player
    Copyright (C) 1999-2002 Masanao Izumo <mo@goice.co.jp>
    Copyright (C) 1995 Tuukka Toivonen <tt@cgs.fi>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; nclude <sys/stat.h>if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    au_a.c

    Functions to output Sun audio file (*.au).
*/

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif /* HAVE_CONFIG_H */
#include <stdio.h>

#ifdef __POCC__
#include <sys/types.h>
#endif //for off_t

#ifdef __W32__
#include <stdlib.h>
#include <io.h>
#endif

#ifdef HAVE_UNISTD_H
#include <unistd.h>
#endif /* HAVE_UNISTD_H */

#ifdef STDC_HEADERS
#include <string.h>
#include <stdlib.h>
#include <ctype.h>
#elif HAVE_STRINGS_H
#include <strings.h>
#endif
#ifdef HAVE_FCNTL_H
#include <fcntl.h>
#endif

#ifdef __FreeBSD__
#include <stdio.h>
#endif

#include "timidity.h"
#include "common.h"
#include "output.h"
#include "controls.h"
#include "instrum.h"
#include "playmidi.h"
#include "readmidi.h"

static int open_output(void); /* 0=success, 1=warning, -1=fatal error */
static void close_output(void);
static int output_data(char *buf, int32 bytes);
static int acntl(int request, void *arg);
static int write_u32(uint32 value);

/* export the playback mode */

#define dpm soundfont_vol_calc

PlayMode dpm = {
    DEFAULT_RATE, PF_PCM_STREAM|PF_FILE_OUTPUT,
	PF_PCM_STREAM|PF_FILE_OUTPUT,
    -1,
    {0,0,0,0,0},
    "Soundfont Volume Calc", 'V',
    NULL,
    open_output,
    close_output,
    output_data,
    acntl
};

long now_max_volume;
int open_output(void)
{
	int include_enc, exclude_enc;
	include_enc = exclude_enc = 0;

	include_enc |= PE_16BIT|PE_SIGNED;
	exclude_enc |= PE_BYTESWAP|PE_24BIT;
	dpm.encoding = validate_encoding(dpm.encoding, include_enc, exclude_enc);

#if !defined ( IA_W32GUI ) && !defined ( IA_W32G_SYN )
	if(dpm.name == NULL) {
		dpm.flag |= PF_AUTO_SPLIT_FILE;
	} else {
		dpm.flag &= ~PF_AUTO_SPLIT_FILE;
	}
#else
	if(w32g_auto_output_mode>0){
		dpm.flag |= PF_AUTO_SPLIT_FILE;
		dpm.name = NULL;
	} else {
		dpm.flag &= ~PF_AUTO_SPLIT_FILE;
	}
#endif
	now_max_volume = 0;
	return 0;
}

static int output_data(char *buf, int32 bytes)
{
    long i=0,tmp=0;
	long maxBytes = bytes>>1;
	short *shb = (short *)buf;

	for (i = 0; i < maxBytes; ++i) {
		tmp = abs( shb[i] );
		if (now_max_volume < tmp) {
			now_max_volume = tmp;
		}
	}
    return 0;
}

static void close_output(void)
{
	FILE *fp = NULL;
	if (dpm.name == NULL)fp = fopen("volume.txt", "w");
	else if (strcmp(dpm.name, "") == 0)fp = fopen("volume.txt", "w");
	else {
		fp = fopen(dpm.name, "w");
		if ((fp = fopen(dpm.name, "w")) == NULL)
			fp = fopen("volume.txt", "w");
	}
	if (fp == NULL)return;
	fprintf(fp, "%.3f\n", 32768.0f / (float)now_max_volume);
	fclose(fp);

	dpm.fd = -1;
}

static int acntl(int request, void *arg)
{
  switch(request) {
  case PM_REQ_PLAY_START:
    if(dpm.flag & PF_AUTO_SPLIT_FILE){
      if(  ( NULL == current_file_info ) || (NULL == current_file_info->filename ) )
        return 0;
      return 0;
   }
    return 0;
  case PM_REQ_PLAY_END:
    if(dpm.flag & PF_AUTO_SPLIT_FILE)
      close_output();
    return 0;
  case PM_REQ_DISCARD:
    return 0;
  }
  return -1;
}
